/* linuxkmsgd: read messages from /proc/kmsg, decode the prefixes and
   write them to stderr, in a suitable format for multilog or splogger
   to deal with.

   Copyright (c) 2000 Adam Sampson, azz@gnu.org. 

   Permission is hereby granted, free of charge, to any person
   obtaining a copy of this software and associated documentation
   files (the "Software"), to deal in the Software without
   restriction, including without limitation the rights to use, copy,
   modify, merge, publish, distribute, sublicense, and/or sell copies
   of the Software, and to permit persons to whom the Software is
   furnished to do so, subject to the following conditions:
  
   The above copyright notice and this permission notice shall be
   included in all copies or substantial portions of the Software.
   
   THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
   EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
   MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
   NONINFRINGEMENT.  IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
   CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF
   CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
   WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/

#include <fcntl.h>
#include <unistd.h>
#include <errno.h>

#define WRITEMSG(msg) { char s[] = msg; write(2, s, (sizeof s) - 1); }

int main(int argc, char **argv) {
  int kmsgfd;
#define LOGBUFSIZE 16384
  char logbuf[LOGBUFSIZE];

  kmsgfd = open("/proc/kmsg", O_RDONLY);
  if (kmsgfd < 0) {
    WRITEMSG("linuxkmsgd: unable to open /proc/kmsg\n");
    return 20;
  }

  while (1) {
    int first = 0, pos = 0, len = 0, bufcount;

    bufcount = read(kmsgfd, logbuf, sizeof(logbuf));
    if (bufcount < 0) {
      if (errno == EINTR) continue;
      WRITEMSG("linuxkmsgd: unable to read from /proc/kmsg\n");
      return 21;
    }

    /* Iterate through the received message, splitting it up into
       lines so that we can decode the loglevel prefixes. */
    while (pos <= bufcount) {

      /* Cope with both \n and \r; comments in the original klogd
	 suggest that some kernels use both. */
      if (logbuf[pos] == '\n' 
	  || logbuf[pos] == '\r'
	  || pos == bufcount) {

	len = pos - first;

	/* We don't bother with zero-length messages. */
	if (len > 0) {

	  /* Decode a <x> prefix if one exists. */
	  if (len > 2 && logbuf[first] == '<' && logbuf[first + 2] == '>') {
	    switch (logbuf[first + 1]) {
	    case '0': WRITEMSG("emerg: "); break;
	    case '1': WRITEMSG("alert: "); break;
	    case '2': WRITEMSG("crit: "); break;
	    case '3': WRITEMSG("err: "); break;
	    case '4': WRITEMSG("warning: "); break;
	    case '5': WRITEMSG("notice: "); break;
	    case '6': WRITEMSG("info: "); break;
	    case '7': WRITEMSG("debug: "); break;
	    default: 
	      /* Unknown <x> prefix -- just display it. */
	      goto noprefix;
	    }
	    len -= 3;
	    first += 3;
	  }
	noprefix:

	  /* Print the message to stderr. */
	  write(2, logbuf + first, len);
	  WRITEMSG("\n");
	}
	first = pos + 1;
      }
      ++pos;
    }
  }

  return 0; /* NOTREACHED */
}
